/// \file
// Range v3 library
//
//  Copyright Eric Niebler 2017-present
//
//  Use, modification and distribution is subject to the
//  Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)
//
// Project home: https://github.com/ericniebler/range-v3
//

#ifndef RANGES_V3_VERSION_HPP
#define RANGES_V3_VERSION_HPP

#define RANGE_V3_MAJOR 0
#define RANGE_V3_MINOR 12
#define RANGE_V3_PATCHLEVEL 0

#define RANGE_V3_VERSION \
    (RANGE_V3_MAJOR * 10000 + RANGE_V3_MINOR * 100 + RANGE_V3_PATCHLEVEL)

#endif
