/// \file
// Range v3 library
//
//  Copyright Eric Niebler 2013,2014.
//
//  Use, modification and distribution is subject to the
//  Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)
//
// Project home: https://github.com/ericniebler/range-v3
//

#ifndef RANGES_V3_ALL_HPP
#define RANGES_V3_ALL_HPP

#include <range/v3/action.hpp>
#include <range/v3/algorithm.hpp>
#include <range/v3/core.hpp>
#include <range/v3/functional.hpp>
#include <range/v3/iterator.hpp>
#include <range/v3/numeric.hpp>
#include <range/v3/range.hpp>
#include <range/v3/utility.hpp>
#include <range/v3/view.hpp>

#endif
